"""Some utilities for generating Markdown."""

from contextlib import contextmanager
from typing import Optional


def generate_markdown_table(
    id: str,
    data: list[tuple[str, float]],
    caption: str,
    short_caption: Optional[str] = None,
) -> str:
    """Generate a Markdown table for a given centrality metric."""
    output: list[str] = []

    output.append(f"\n<!-- BEGIN {id} -->")
    output.append(
        f"table: {caption} {{#tbl:{id}{(' shortcaption="' + short_caption + '"') if short_caption else ''}}}\n"
    )
    output.append("| Package | Score |")
    output.append(f"|{FULL_WIDTH_HEADER_SEPARATOR}|------:|")
    for i, (pkg, score) in enumerate(data, 1):
        output.append(f"| {table_margin_marker(i)}`{pkg}` | {score:.4f} |")
    output.append(f"\n<!-- END {id} -->")

    return "\n".join(output)


@contextmanager
def surround_with_marker(marker: str):
    try:
        print(f"\n<!-- BEGIN {marker} -->\n")
        yield marker
    finally:
        print(f"\n<!-- END {marker} -->\n")


FULL_WIDTH_HEADER_SEPARATOR = "-" * 80


def table_margin_marker(marker: int):
    return f"\\marginnote{{\\scriptsize{{  {marker} }}}}" if marker % 5 == 0 else ""



from pathlib import Path
from typing import Optional
import networkx as nx

NODE_COLOR = "#b8d5b8"
ROOT_COLOR = "#ce6c47"

# TODO:
# - Colour the nodes based on out-degree
# - Perhaps colour based on the different analyses?


def generate_dot_data(graph: nx.DiGraph) -> list[str]:
    """
    Generate DOT data for the given dependency graph.

    Args:
        graph: The dependency graph to export.
    """
    graph_to_export = graph

    # Start the DOT file
    dot_data = [
        "digraph DependencyNetwork {",
    ]

    graph_settings = [
        "  graph [",
        "    layout=dot,",  # Use dot layout engine
        "    rankdir=LR,",  # Left to right layout
        "    ranksep=1.0",  # Rank separation
        "    pad=0.5",  # Add padding around the graph (in inches)
        # '    bgcolor="transparent"',  # Set transparent background
        "  ];",
    ]

    node_settings = [
        "  node [",
        "    shape=box,",  # Box shape for nodes
        "    style=filled,",  # Fill nodes with color
        "    fontsize=10",  # Font size
        "  ];",
    ]

    edge_settings = [
        "  edge [",
        "    arrowsize=0.4,",  # Smaller arrowheads
        "    arrowhead=vee",  # Optional: use vee style arrowheads which look cleaner
        '    color="#888888"',  # Grey color for edges
        "    penwidth=0.5,",  # Thinner lines
        "    style=bezier",  # Use bezier curves for smoother edges
        "  ];",
    ]

    edge_handling = [
        "  concentrate=true;",  # Merge edges going to the same destination
        # "  compound=true;",  # Enable edge routing between clusters
        "  mclimit=1.5;",  # Increase iterations for edge crossing minimization
    ]

    # Add all settings to the DOT data
    dot_data.extend(graph_settings)
    dot_data.extend(node_settings)
    dot_data.extend(edge_settings)
    dot_data.extend(edge_handling)

    # Add nodes with attributes
    for node, data in graph_to_export.nodes(data=True):
        node_name = data.get("name", node.split("@")[0])
        node_type = data.get("type", "unknown")

        # Set color based on node type
        if node_type == "root":
            color = ROOT_COLOR
        elif node_type == "dependency":
            color = NODE_COLOR
        else:
            color = "white"

        # Create a label with package name and version
        if "version" in data:
            label = f"{node_name}\\n{data['version']}"
        elif "version_constraint" in data:
            label = f"{node_name}\\n{data['version_constraint']}"
        else:
            label = node_name

        # Escape quotes in node names for DOT format
        safe_node = node.replace('"', '\\"')

        if node_type == "root":
            dot_data.append(
                f'  "{safe_node}" [label="{label}", fillcolor="{color}", style="filled, bold", fontsize=14, padding=4, width=1.5, height=6, penwidth=2];'
            )
        else:
            dot_data.append(f'  "{safe_node}" [label="{label}", fillcolor="{color}"];')

    # Add edges with attributes
    for source, target, data in graph_to_export.edges(data=True):
        edge_type = data.get("type", "unknown")

        # Set edge style based on type
        if edge_type == "installed":
            style = "solid"
        elif edge_type == "required":
            style = "dashed"
        else:
            style = "dotted"

        # Escape quotes in node names for DOT format
        safe_source = source.replace('"', '\\"')
        safe_target = target.replace('"', '\\"')

        dot_data.append(f'  "{safe_source}" -> "{safe_target}" [style={style}];')

    dot_data.append("}")
    return dot_data


def export_to_dot(graph: nx.DiGraph, output_file: Optional[Path]):
    """
    Export the dependency graph to a DOT file.

    Args:
        graph: The dependency graph to export.
        output_file: The path to the output DOT file.
    """
    dot_data = "\n".join(generate_dot_data(graph))
    if not output_file:
        print(dot_data)
        return

    # Write to file
    with open(output_file, "w") as f:
        f.write(dot_data)
        print(f"DOT file exported to {output_file}")



from dataclasses import dataclass
from typing import Iterable, Optional
import networkx as nx
from networkx import find_cycle
import numpy as np

from plockalyser.markdown import generate_markdown_table, table_margin_marker

TOP_X_TO_SHOW = 20
BOTTOM_X_TO_SHOW = 10


def analyse_connectivity(graph: nx.DiGraph) -> str:
    """Analyse the degree of connectivity in the network."""
    assert callable(graph.degree)
    degrees = dict(graph.degree())
    in_degrees = dict(graph.in_degree())
    out_degrees = dict(graph.out_degree())

    # Sort by total degree
    sorted_degrees = sorted(degrees.items(), key=lambda x: x[1], reverse=True)

    output: list[str] = []
    marker = "degree_of_connectivity"
    output.append(f"\n<!-- BEGIN {marker} -->")

    output.append(f"table: Degree of connectivity per package {{#tbl:{marker}}}\n")
    output.append("| Package | In | Out | Tot. |")
    # First column really wide to make the table full-width with Pandoc/LaTeX.
    output.append(
        "|--------------------------------------------------------------------|-----:|-----:|-----:|"
    )
    for i, (pkg, degree) in enumerate(sorted_degrees[:TOP_X_TO_SHOW], 1):
        output.append(
            f"| {table_margin_marker(i)}`{pkg}` | {in_degrees[pkg]} | {out_degrees[pkg]} | {degree} |"
        )
    output.append(f"\n<!-- END {marker} -->")
    return "\n".join(output)


def analyse_closeness(graph: nx.DiGraph) -> str:
    closeness_in = nx.closeness_centrality(graph)
    # reverse to get the out links (the dependencies)
    closeness_out = nx.closeness_centrality(graph.reverse())
    sorted_closeness_in = sorted(closeness_in.items(), key=lambda x: x[1], reverse=True)
    sorted_closeness_out = sorted(
        closeness_out.items(), key=lambda x: x[1], reverse=True
    )

    output: list[str] = []
    output.append(
        generate_markdown_table(
            "closeness_centrality_in",
            sorted_closeness_in[:TOP_X_TO_SHOW],
            "Closeness centrality -- dependents, highest",
        )
    )

    output.append(
        generate_markdown_table(
            "closeness_centrality_in_lowest",
            sorted_closeness_in[-BOTTOM_X_TO_SHOW:],
            "Closeness centrality -- dependents, lowest",
        )
    )

    output.append(
        generate_markdown_table(
            "closeness_centrality_out",
            sorted_closeness_out[:TOP_X_TO_SHOW],
            "Closeness centrality -- dependencies, highest",
        )
    )

    output.append(
        generate_markdown_table(
            "closeness_centrality_out_lowest",
            [(p, s) for p, s in sorted_closeness_out if s > 0][-BOTTOM_X_TO_SHOW:],
            "Closeness centrality -- dependencies, lowest non-zero",
        )
    )

    return "\n".join(output)


def analyse_betweenness(graph: nx.DiGraph) -> str:
    betweenness_directed = nx.betweenness_centrality(graph)
    betweenness_undirected = nx.betweenness_centrality(graph.to_undirected())
    sorted_betweenness_directed = sorted(
        betweenness_directed.items(), key=lambda x: x[1], reverse=True
    )
    sorted_betweenness_undirected = sorted(
        betweenness_undirected.items(), key=lambda x: x[1], reverse=True
    )

    output: list[str] = []

    output.append(
        generate_markdown_table(
            "betweenness_centrality_directed",
            sorted_betweenness_directed[:TOP_X_TO_SHOW],
            "Betweenness centrality -- directed",
        )
    )

    output.append(
        generate_markdown_table(
            "betweenness_centrality_undirected",
            sorted_betweenness_undirected[:TOP_X_TO_SHOW],
            "Betweenness centrality -- undirected",
        )
    )

    return "\n".join(output)


def analyse_pagerank(graph: nx.DiGraph) -> str:
    pagerank = nx.pagerank(graph)  # PageRank as a prestige metric
    sorted_pagerank = sorted(pagerank.items(), key=lambda x: x[1], reverse=True)

    output: list[str] = []
    output.append(
        generate_markdown_table(
            "pagerank_centrality",
            sorted_pagerank[:TOP_X_TO_SHOW],
            "Prestige centrality (PageRank algorithm)",
        )
    )
    return "\n".join(output)


def analyse_centrality(graph: nx.DiGraph) -> str:
    output: list[str] = []
    output.append(analyse_closeness(graph))
    output.append(analyse_betweenness(graph))
    output.append(analyse_pagerank(graph))
    return "\n\n".join(output)


@dataclass
class DependencyNetworkStats:
    """Statistics about a dependency network."""

    nodes: int
    edges: int
    direct_dependencies: Optional[int]
    max_path_length: int
    avg_path_length: float
    clustering: float
    density: float
    packages_with_more_than_one_version: int


def get_summary_statistics(graph: nx.DiGraph) -> DependencyNetworkStats:
    """Get summary statistics about the dependency network."""
    # Basic stats
    nodes = graph.number_of_nodes()
    edges = graph.number_of_edges()

    # Find which packages that have more than one version:
    version_count: dict[str, int] = {}
    for node, nodedata in graph.nodes.items():
        if nodedata["name"] in version_count:
            version_count[nodedata["name"]] += 1
        else:
            version_count[nodedata["name"]] = 1
    more_than_one_version = {k: v for k, v in version_count.items() if v > 1}

    # Find loops. Just print them out for now.
    loops = find_cycle(graph)
    print(f"{loops=}")

    # Find the root node (should have in_degree of 0)
    root_nodes = [node for node in graph.nodes() if graph.in_degree(node) == 0]
    assert len(root_nodes) == 1
    root_node = root_nodes[0]  # Assume the first root node is the main one
    # Count direct dependencies (distance 1 from root)
    direct_deps = list(graph.successors(root_node))
    direct_dependencies = len(direct_deps)

    # Connected components in the undirected graph
    undirected = graph.to_undirected()
    connected_components = list(nx.connected_components(undirected))

    # Average path length (only for the largest component to avoid errors)
    largest_cc = max(connected_components, key=len)
    largest_subgraph = undirected.subgraph(largest_cc)
    max_path_length = nx.diameter(largest_subgraph)
    avg_path_length = nx.average_shortest_path_length(largest_subgraph)

    # Clustering coefficient
    clustering = nx.average_clustering(undirected)

    # Density
    density = nx.density(graph)

    return DependencyNetworkStats(
        nodes=nodes,
        edges=edges,
        direct_dependencies=direct_dependencies,
        max_path_length=max_path_length,
        avg_path_length=avg_path_length,
        clustering=clustering,
        density=density,
        packages_with_more_than_one_version=len(more_than_one_version),
    )


def format_summary_statistics(stats: DependencyNetworkStats) -> str:
    """Print summary statistics about the dependency network."""

    output: list[str] = []
    marker = "basic_stats"

    # Basic stats
    output.append(f"\n<!-- BEGIN {marker} -->\n")
    output.append("table: Basic network statistics {#tbl:basic_stats}\n")
    output.append("| Basic statistics | Value |")
    output.append("|-------|-------:|")
    output.append(f"| Number of packages (nodes) | {stats.nodes} |")
    output.append(f"| Number of dependencies (edges) | {stats.edges} |")
    # Direct dependencies
    if stats.direct_dependencies is not None:
        output.append(
            f"| Number of direct dependencies of `root` | {stats.direct_dependencies} |"
        )
    else:
        output.append(
            "| Number of direct dependencies of `root` | Warning: No `root` node found |"
        )
    output.append(f"| Maximum path length | {stats.max_path_length} |")
    output.append(
        f"| Packages with more than one version | {stats.packages_with_more_than_one_version} |"
    )
    output.append(f"\n<!-- END {marker} -->\n")
    return "\n".join(output)


def calculate_gini_coefficients(data: Iterable[float]) -> float:
    """Calculate the Gini coefficient for the data.

    The Gini coefficient measures inequality in a distribution.
    A value of 0 represents perfect equality, while 1 represents maximum inequality.
    """
    values = sorted(data)
    n = len(values)

    # If there are no values or only one value, return 0 (perfect equality)
    if n <= 1:
        return 0.0

    # Slightly adapted from https://stackoverflow.com/a/61154922

    x = np.array(data)
    diffsum: float = 0
    for i, xi in enumerate(x[:-1], 1):
        diffsum += np.sum(np.abs(xi - x[i:]))

    return diffsum / (len(x) ** 2 * np.mean(x, dtype=np.float64))


def format_gini(graph: nx.DiGraph) -> str:
    """Print Gini coefficients for some of the graph's measures."""

    data: list[tuple[str, float]] = []

    # Closeness (dependencies)
    closeness_out = nx.closeness_centrality(graph.reverse())
    gini_closeness_out = calculate_gini_coefficients(
        [v for v in closeness_out.values()]
    )
    data.append(("Closeness (dependencies)", gini_closeness_out))

    # Degree of connectivity (out-degree/dependencies)
    out_degree = graph.out_degree()
    assert type(out_degree) is not int
    gini_out_degree = calculate_gini_coefficients([v for _, v in out_degree])
    data.append(("Degree of connectivity (dependencies)", gini_out_degree))

    # Prestige (PageRank)
    pagerank = nx.pagerank(graph)
    gini_pagerank = calculate_gini_coefficients([v for v in pagerank.values()])
    data.append(("Prestige (PageRank)", gini_pagerank))

    id = "gini_coefficients"
    caption = "Gini coefficients"

    output: list[str] = []

    output.append(f"\n<!-- BEGIN {id} -->")
    output.append(f"table: {caption} {{#tbl:{id}}}\n")
    output.append("| Measure | Gini coefficient |")
    # output.append(f"|{FULL_WIDTH_HEADER_SEPARATOR}|------:|")
    output.append("|------|------:|")
    for i, (pkg, score) in enumerate(data, 1):
        output.append(f"| {pkg} | {score:.4f} |")
    output.append(f"\n<!-- END {id} -->")

    return "\n".join(output)



#!.venv/bin/python3
import json
from pathlib import Path
from typing import Optional
import networkx as nx
import argparse
from pydantic import BaseModel

from plockalyser.analyses import (
    analyse_centrality,
    analyse_connectivity,
    get_summary_statistics,
    format_summary_statistics,
    format_gini,
)
from plockalyser.dot import export_to_dot


class Dependency(BaseModel):
    model_config = {"arbitrary_types_allowed": True}
    version: Optional[str] = None
    resolved: Optional[str] = None
    overridden: Optional[bool] = None
    dependencies: Optional[dict[str, "Dependency"]] = None


class NpmLsError(BaseModel):
    code: str
    summary: str
    detail: str


class NpmLsOutput(BaseModel):
    version: str
    name: str
    problems: list[str]
    dependencies: dict[str, Dependency]


def load_and_parse(package_lock_path: Path) -> tuple[nx.DiGraph, NpmLsOutput]:
    """Load the npm ls output file and build a dependency graph."""
    with open(package_lock_path, "r") as f:
        data = json.load(f)

    data = NpmLsOutput(**data)

    root_name = "root"  # data.name
    root_version = ""  # data.version
    # root_id = f"{root_name}@{root_version}"
    root_id = f"{root_name}"
    graph = nx.DiGraph()

    # Add the root node
    graph.add_node(root_id, name=root_name, version=root_version, type="root")

    # Process dependencies at the top level
    _process_dependencies_tree(graph, data.dependencies, root_id)
    return (graph, data)


def _process_dependencies_tree(
    graph: nx.DiGraph, dependencies: dict[str, Dependency], parent_id: str
):
    for pkg_name, pkg_info in dependencies.items():
        version = pkg_info.version or "unknown"
        pkg_id = f"{pkg_name}@{version}"

        # Add the node if it doesn't exist
        if not graph.has_node(pkg_id):
            graph.add_node(
                pkg_id,
                name=pkg_name,
                version=version,
                type="dependency",
            )

        # Add the edge from parent to this dependency (installed dependency)
        graph.add_edge(parent_id, pkg_id, type="installed")

        # Recursively process nested dependencies
        if pkg_info.dependencies:
            _process_dependencies_tree(graph, pkg_info.dependencies, pkg_id)


def output_tables(tables: list[str], output_file: Optional[Path]):
    if not output_file:
        print("\n\n".join(tables))
        return

    with open(output_file, "w") as f:
        f.write("\n\n".join(tables))
        print(f"Wrote tables to {output_file}")


def main():
    parser = argparse.ArgumentParser(
        description="Analyse a dependency network generated by `npm ls --all --json`"
    )
    parser.add_argument("npm_ls_file", help="Path to package-lock.json file")
    parser.add_argument(
        "--tables",
        help="Export network analyses as Markdown tables",
        action="store_true",
    )
    parser.add_argument(
        "--dot", help="Export network to DOT format", action="store_true"
    )
    parser.add_argument("--output", help="File to output to", default=None)

    args = parser.parse_args()

    graph, _ = load_and_parse(args.npm_ls_file)

    if args.output and sum([args.dot, args.tables]) > 1:
        print("When --output is given, --dot and --tables are mutually exclusive.")
        return 1

    if args.dot:
        export_to_dot(graph, args.output)

    # Run analyses
    if args.tables:
        tables_output: list[str] = []
        tables_output.append(format_summary_statistics(get_summary_statistics(graph)))
        tables_output.append(analyse_connectivity(graph))
        tables_output.append(analyse_centrality(graph))
        tables_output.append(format_gini(graph))
        output_tables(tables_output, args.output)


if __name__ == "__main__":
    main()
